"""
         Principal Components Analysis
              Yasuharu Okamoto
                  
"""
import pandas as pd
import numpy as np
import scipy.stats as ss
import matplotlib.pyplot as plt

fl_nm = input('Excel Data File(*.xlsx) = ')
xlfile = pd.ExcelFile(fl_nm)
df = pd.read_excel(xlfile)
fout_nm = input('Output Data file(*.txt) = ')
fout = open(fout_nm, 'w')
fout.write(f'Data file = {fl_nm}\n')

genvars = np.array(df.keys())[1:]
print('genvars =\n', genvars)
case_id = df.values.T[0]
n = len(case_id)
print('case_id =\n', case_id)
genX = np.array(df.values[:, 1:], dtype = float)
print('genX =\n', genX)

fout.write('\nData =\n')
for i in range(n):
    fout.write(f'{case_id[i]:>20}')
    for v in genX[i]:
        fout.write(f'{v:>10.3f}')
    fout.write('\n')
fout.write('\n')

means = np.mean(genX, axis=0)
sds = np.std(genX, axis=0, ddof=0)
fout.write(f'\nmeans =\n{means}')
fout.write(f'\nsds =\n{sds}\n')

genX = (genX - means) / sds        #   Standardization

fout.write('\n\nStandardized Data =\n')
for i in range(n):
    fout.write(f'{case_id[i]:>20}')
    for v in genX[i]:
        fout.write(f'{v:>10.3f}')
    fout.write('\n')
fout.write('\n')

#
#        Principal Components Analysis
#   式（5.2.2）, 岡本安晴（2014）心理学データ分析と測定．勁草書房
#

U, S, Vh = np.linalg.svd(genX/(n**0.5))

print('\nSingular values...\n', S)
fout.write(f'\nSingular values...\n{S}\n')

plt.plot(range(len(S)), S)
plt.ylabel('Singular value')
plt.xticks(range(len(S)), np.arange(len(S))+1)
plt.show()
n_comp = int(input('Number of components = '))
rnk_S = np.sum([S > 1.0e-9])
print('\nrnk_S =', rnk_S)
print('\nRank of genX =', rnk_S)
fout.write(f'\nRank of genX = {rnk_S}\n')
if rnk_S < n_comp:
    n_comp = rnk_S
print('\nn_comp =', n_comp)
fout.write(f'\nn_comp = {n_comp}\n')

C = U * (n**0.5)          #   Principal Components

print('\nPrincipal components...')
fout.write('\nPrincipal components...\n')
print(f'{" ":>15s}', end = '')
fout.write(f'{" ":>15s}')
for i in range(n_comp):
    v = f'comp_{i+1}'
    print(f'{v:>15s}', end = '')
    fout.write(f'{v:>15s}')
print()
fout.write('\n')
for i in range(len(case_id)):
    print(f'{case_id[i]:>15s}', end = '')
    fout.write(f'{case_id[i]:>15s}')
    for j in range(n_comp):
        print(f'{C[i][j]:>15.5f}', end = '')
        fout.write(f'{C[i][j]:>15.5f}')
    print()
    fout.write('\n')
      
max_c = np.max(C)
min_c = np.min(C)
if max_c < -min_c:
    max_c = -min_c
if min_c > -max_c:
    min_c = -max_c
dimx = 1
dimy = 2 if n_comp > 1 else 1
while True:
    for i in range(len(case_id)):
        plt.plot(C[i][dimx-1], C[i][dimy-1], 'o', c = 'b')
        plt.text(C[i][dimx-1], C[i][dimy-1], case_id[i])
    plt.xlabel(f'comp_{dimx}')
    plt.ylabel(f'comp_{dimy}')
    plt.plot([min_c, max_c], [0,0], lw = 1, c ='k', ls = ':')
    plt.plot([0,0], [min_c, max_c], lw = 1, c = 'k', ls = ':')
    plt.show()
    dimx = int(input('X-Axis = '))
    if dimx < 1:
        break
    if dimx > n_comp:
        dimx = n_comp
    dimy = int(input('Y-Axis = '))
    if dimy < 1:
        break
    if dimy > n_comp:
        dimy = n_comp


Strctr = genX.T @ C / n        #   Structure Matrix
    
dimx = 1
dimy = 2 if n_comp > 1 else 1
while True:
    for i in range(len(genX[0])):
        plt.plot(Strctr[i][dimx-1], Strctr[i][dimy-1], 'o', c = 'b')
        plt.text(Strctr[i][dimx-1], Strctr[i][dimy-1], genvars[i])
    plt.plot([-1, 1], [0,0], lw = 1, c ='k', ls = ':')
    plt.plot([0,0], [-1, 1], lw = 1, c = 'k', ls = ':')
    plt.xlabel(f'comp_{dimx}')
    plt.ylabel(f'comp_{dimy}')
    plt.show()
    dimx = int(input('X-Axis = '))
    if dimx < 1:
        break
    if dimx > n_comp:
        dimx = n_comp
    dimy = int(input('Y-Axis = '))
    if dimy < 1:
        break
    if dimy > n_comp:
        dimy = n_comp

print('\nStructure matrix...')
fout.write('\nStructure matrix...\n')
print(f'{" ":>15s}', end = '')
fout.write(f'{" ":>15s}')
for i in range(n_comp):
    v = f'comp_{i+1}'
    print(f'{v:>15s}', end = '')
    fout.write(f'{v:>15s}')
print()
fout.write('\n')
for i in range(len(genvars)):
    print(f'{genvars[i]:>15s}', end = '')
    fout.write(f'{genvars[i]:>15s}')
    for j in range(n_comp):
        print(f'{Strctr[i][j]:>15.5f}', end = '')
        fout.write(f'{Strctr[i][j]:15.5f}')
    print()
    fout.write('\n')
    
fout.close()
print(f'\n{fout_nm} was saved.\n')

fout.close()
